package ru.dbotthepony.mc.otm.core.math

import net.minecraft.core.BlockPos
import net.minecraft.core.Direction
import net.minecraft.core.Vec3i
import net.minecraft.util.StringRepresentable
import net.minecraft.world.level.block.Rotation
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.capabilities.ICapabilityProvider
import net.minecraftforge.common.util.LazyOptional

internal inline val Direction.blockRotation
	get() = BlockRotation.of(this)

fun <T> ICapabilityProvider.getCapability(capability: Capability<T>, side: BlockRotation?): LazyOptional<T> {
	return getCapability(capability, side?.primary)
}

operator fun Vec3i.plus(other: BlockRotation): Vec3i {
	return this + other.normal
}

operator fun BlockPos.plus(other: BlockRotation): BlockPos {
	return this + other.normal
}

enum class BlockRotation(val primary: Direction, val secondary: Direction?) : StringRepresentable {
	DOWN(Direction.DOWN, null),
	UP(Direction.UP, null),
	NORTH(Direction.NORTH, null),
	SOUTH(Direction.SOUTH, null),
	WEST(Direction.WEST, null),
	EAST(Direction.EAST, null),

	DOWN_NORTH(Direction.DOWN, Direction.NORTH),
	DOWN_SOUTH(Direction.DOWN, Direction.SOUTH),
	DOWN_WEST(Direction.DOWN, Direction.WEST),
	DOWN_EAST(Direction.DOWN, Direction.EAST),

	UP_NORTH(Direction.UP, Direction.NORTH),
	UP_SOUTH(Direction.UP, Direction.SOUTH),
	UP_WEST(Direction.UP, Direction.WEST),
	UP_EAST(Direction.UP, Direction.EAST);

	val lowercaseName = name.lowercase()

	override fun getSerializedName(): String {
		return lowercaseName
	}

	fun rotate(by: Rotation): BlockRotation {
		return of(by.rotate(primary), secondary)
	}

	operator fun times(other: Int): Vec3i {
		return normal * other
	}

	operator fun plus(other: Vec3i): Vec3i {
		return normal + other
	}

	operator fun plus(other: Direction): Vec3i {
		return normal + other
	}

	operator fun plus(other: BlockRotation): Vec3i {
		return normal + other.normal
	}

	operator fun unaryMinus() = opposite

	// more performant
	val opposite by lazy { of(primary.opposite, secondary?.opposite) }
	val normal: Vec3i get() = primary.normal

	companion object {
		@JvmStatic
		fun of(direction: Direction): BlockRotation {
			return when (direction) {
				Direction.DOWN -> DOWN
				Direction.UP -> UP
				Direction.NORTH -> NORTH
				Direction.SOUTH -> SOUTH
				Direction.WEST -> WEST
				Direction.EAST -> EAST
			}
		}

		@JvmStatic
		fun of(primary: Direction, secondary: Direction?): BlockRotation {
			return when (primary) {
				Direction.NORTH -> {
					require(secondary == null) { "Impossible direction: $primary - $secondary" }
					NORTH
				}

				Direction.SOUTH -> {
					require(secondary == null) { "Impossible direction: $primary - $secondary" }
					SOUTH
				}

				Direction.WEST -> {
					require(secondary == null) { "Impossible direction: $primary - $secondary" }
					WEST
				}

				Direction.EAST -> {
					require(secondary == null) { "Impossible direction: $primary - $secondary" }
					EAST
				}

				Direction.UP -> {
					when (secondary) {
						null -> UP
						Direction.NORTH -> UP_NORTH
						Direction.SOUTH -> UP_SOUTH
						Direction.WEST -> UP_WEST
						Direction.EAST -> UP_EAST
						else -> throw IllegalArgumentException("Impossible direction: $primary - $secondary")
					}
				}

				Direction.DOWN -> {
					when (secondary) {
						null -> DOWN
						Direction.NORTH -> DOWN_NORTH
						Direction.SOUTH -> DOWN_SOUTH
						Direction.WEST -> DOWN_WEST
						Direction.EAST -> DOWN_EAST
						else -> throw IllegalArgumentException("Impossible direction: $primary - $secondary")
					}
				}
			}
		}
	}
}
